@@ -1,3 +1,13 @@
+###################################################################
+#  Numexpr - Fast numerical array expression evaluator for NumPy.
+#
+#      License: MIT
+#      Author:  See AUTHORS.txt
+#
+#  See LICENSE.txt and LICENSES/*.txt for details about copyright and
+#  rights to use.
+####################################################################
+
 """
 Numexpr is a fast numerical expression evaluator for NumPy.  With it,
 expressions that operate on arrays (like "3*a+4*b") are accelerated
@@ -34,6 +44,9 @@ from numexpr.utils import (
 
 # Initialize the number of threads to be used
 ncores = detect_number_of_cores()
+# Check that we don't surpass the MAX_THREADS in interpreter.c
+if ncores > 4096:
+    ncores = 4096
 set_num_threads(ncores)
 # The default for VML is 1 thread (see #39)
 set_vml_num_threads(1)
