@@ -1,3 +1,13 @@
+###################################################################
+#  Numexpr - Fast numerical array expression evaluator for NumPy.
+#
+#      License: MIT
+#      Author:  See AUTHORS.txt
+#
+#  See LICENSE.txt and LICENSES/*.txt for details about copyright and
+#  rights to use.
+####################################################################
+
 __all__ = ['E']
 
 import operator
@@ -60,7 +70,7 @@ def ophelper(f):
             if isConstant(x):
                 args[i] = x = ConstantNode(x)
             if not isinstance(x, ExpressionNode):
-                raise TypeError("unsupported object type: %s" % (type(x),))
+                raise TypeError("unsupported object type: %s" % type(x))
         return f(*args)
     func.__name__ = f.__name__
     func.__doc__ = f.__doc__
@@ -92,6 +102,7 @@ def commonKind(nodes):
 
 max_int32 = 2147483647
 min_int32 = -max_int32 - 1
+
 def bestConstantType(x):
     if isinstance(x, str):  # ``numpy.string_`` is a subclass of ``str``
         return str
@@ -103,7 +114,7 @@ def bestConstantType(x):
     # ``double`` objects are kept as is to allow the user to force
     # promotion of results by using double constants, e.g. by operating
     # a float (32-bit) array with a double (64-bit) constant.
-    if isinstance(x, (double)):
+    if isinstance(x, double):
         return double
     # Numeric conversion to boolean values is not tried because
     # ``bool(1) == True`` (same for 0 and False), so 0 and 1 would be
@@ -111,23 +122,24 @@ def bestConstantType(x):
     # supported.
     if isinstance(x, (bool, numpy.bool_)):
         return bool
-    # ``long`` is not explicitly needed since ``int`` automatically
-    # returns longs when needed (since Python 2.3).
+    if isinstance(x, (int, numpy.integer)):
+        # Constants needing more than 32 bits are always
+        # considered ``long``, *regardless of the platform*, so we
+        # can clearly tell 32- and 64-bit constants apart.
+        if not (min_int32 <= x <= max_int32):
+            return long
+        return int
     # The duality of float and double in Python avoids that we have to list
     # ``double`` too.
-    for converter in int, float, complex:
+    for converter in float, complex:
         try:
             y = converter(x)
         except StandardError, err:
             continue
-        if x == y:
-            # Constants needing more than 32 bits are always
-            # considered ``long``, *regardless of the platform*, so we
-            # can clearly tell 32- and 64-bit constants apart.
-            if converter is int and not (min_int32 <= x <= max_int32):
-                return long
+        if y == x:
             return converter
 
+
 def getKind(x):
     converter = bestConstantType(x)
     return type_to_kind[converter]
@@ -154,6 +166,9 @@ def func(func, minkind=None, maxkind=None):
         kind = commonKind(args)
         if kind in ('int', 'long'):
             # Exception for following NumPy casting rules
+            #FIXME: this is not always desirable. The following
+            # functions which return ints (for int inputs) on numpy
+            # but not on numexpr: copy, abs, fmod, ones_like
             kind = 'double'
         else:
             # Apply regular casting rules
@@ -167,6 +182,7 @@ def func(func, minkind=None, maxkind=None):
 @ophelper
 def where_func(a, b, c):
     if isinstance(a, ConstantNode):
+        #FIXME: This prevents where(True, a, b)
         raise ValueError("too many dimensions")
     if allConstantNodes([a,b,c]):
         return ConstantNode(numpy.where(a, b, c))
@@ -179,12 +195,12 @@ def encode_axis(axis):
         axis = interpreter.allaxes
     else:
         if axis < 0:
-            axis = interpreter.maxdims - axis
+            raise ValueError("negative axis are not supported")
         if axis > 254:
             raise ValueError("cannot encode axis")
     return RawNode(axis)
 
-def sum_func(a, axis=-1):
+def sum_func(a, axis=None):
     axis = encode_axis(axis)
     if isinstance(a, ConstantNode):
         return a
@@ -192,7 +208,7 @@ def sum_func(a, axis=-1):
         a = ConstantNode(a)
     return FuncNode('sum', [a, axis], kind=a.astKind)
 
-def prod_func(a, axis=-1):
+def prod_func(a, axis=None):
     axis = encode_axis(axis)
     if isinstance(a, (bool, int, long, float, double, complex)):
         a = ConstantNode(a)
@@ -209,9 +225,25 @@ def div_op(a, b):
             return OpNode('mul', [a, ConstantNode(1./b.value)])
     return OpNode('div', [a,b])
 
+@ophelper
+def truediv_op(a, b):
+    if get_optimization() in ('moderate', 'aggressive'):
+        if (isinstance(b, ConstantNode) and
+            (a.astKind == b.astKind) and
+            a.astKind in ('float', 'double', 'complex')):
+            return OpNode('mul', [a, ConstantNode(1./b.value)])
+    kind = commonKind([a, b])
+    if kind in ('bool', 'int', 'long'):
+        kind = 'double'
+    return OpNode('div', [a, b], kind=kind)
+
+@ophelper
+def rtruediv_op(a, b):
+    return truediv_op(b, a)
+
 @ophelper
 def pow_op(a, b):
-    if allConstantNodes([a,b]):
+    if allConstantNodes([a, b]):
         return ConstantNode(a**b)
     if isinstance(b, ConstantNode):
         x = b.value
@@ -237,7 +269,8 @@ def pow_op(a, b):
                     p = OpNode('mul', [p,p])
                 if ishalfpower:
                     kind = commonKind([a])
-                    if kind in ('int', 'long'): kind = 'double'
+                    if kind in ('int', 'long'):
+                        kind = 'double'
                     r = multiply(r, OpNode('sqrt', [a], kind))
                 if r is None:
                     r = OpNode('ones_like', [a])
@@ -248,7 +281,7 @@ def pow_op(a, b):
             if x == -1:
                 return OpNode('div', [ConstantNode(1),a])
             if x == 0:
-                return FuncNode('ones_like', [a])
+                return OpNode('ones_like', [a])
             if x == 0.5:
                 kind = a.astKind
                 if kind in ('int', 'long'): kind = 'double'
@@ -345,12 +378,22 @@ class ExpressionNode(object):
     def __pos__(self):
         return self
 
+    # The next check is commented out. See #24 for more info.
+
+    # def __nonzero__(self):
+    #     raise TypeError("You can't use Python's standard boolean operators in "
+    #                     "NumExpr expressions. You should use their bitwise "
+    #                     "counterparts instead: '&' instead of 'and', "
+    #                     "'|' instead of 'or', and '~' instead of 'not'.")
+
     __add__ = __radd__ = binop('add')
     __sub__ = binop('sub')
     __rsub__ = binop('sub', reversed=True)
     __mul__ = __rmul__ = binop('mul')
-    __div__ =  div_op
+    __div__ = div_op
     __rdiv__ = binop('div', reversed=True)
+    __truediv__ = truediv_op
+    __rtruediv__ = rtruediv_op
     __pow__ = pow_op
     __rpow__ = binop('pow', reversed=True)
     __mod__ = binop('mod')
