@@ -1,7 +1,25 @@
+/*********************************************************************
+  Numexpr - Fast numerical array expression evaluator for NumPy.
+
+      License: MIT
+      Author:  See AUTHORS.txt
+
+  See LICENSE.txt for details about copyright and rights to use.
+**********************************************************************/
+
+
 {
 #define VEC_LOOP(expr) for(j = 0; j < BLOCK_SIZE; j++) {       \
         expr;                                   \
     }
+
+#define VEC_ARG0(expr)                          \
+    BOUNDS_CHECK(store_in);                     \
+    {                                           \
+        char *dest = mem[store_in];             \
+        VEC_LOOP(expr);                         \
+    } break
+
 #define VEC_ARG1(expr)                          \
     BOUNDS_CHECK(store_in);                     \
     BOUNDS_CHECK(arg1);                         \
@@ -9,7 +27,7 @@
         char *dest = mem[store_in];             \
         char *x1 = mem[arg1];                   \
         intp ss1 = params.memsizes[arg1];       \
-        intp sb1 = params.memsteps[arg1];       \
+        intp sb1 = memsteps[arg1];              \
         /* nowarns is defined and used so as to \
         avoid compiler warnings about unused    \
         variables */                            \
@@ -17,6 +35,7 @@
         nowarns += 1;                           \
         VEC_LOOP(expr);                         \
     } break
+
 #define VEC_ARG2(expr)                          \
     BOUNDS_CHECK(store_in);                     \
     BOUNDS_CHECK(arg1);                         \
@@ -25,14 +44,14 @@
         char *dest = mem[store_in];             \
         char *x1 = mem[arg1];                   \
         intp ss1 = params.memsizes[arg1];       \
-        intp sb1 = params.memsteps[arg1];       \
+        intp sb1 = memsteps[arg1];              \
         /* nowarns is defined and used so as to \
         avoid compiler warnings about unused    \
         variables */                            \
         intp nowarns = ss1+sb1+*x1;             \
         char *x2 = mem[arg2];                   \
         intp ss2 = params.memsizes[arg2];       \
-        intp sb2 = params.memsteps[arg2];       \
+        intp sb2 = memsteps[arg2];              \
         nowarns += ss2+sb2+*x2;                 \
         VEC_LOOP(expr);                         \
     } break
@@ -46,17 +65,17 @@
         char *dest = mem[store_in];             \
         char *x1 = mem[arg1];                   \
         intp ss1 = params.memsizes[arg1];       \
-        intp sb1 = params.memsteps[arg1];       \
+        intp sb1 = memsteps[arg1];              \
         /* nowarns is defined and used so as to \
         avoid compiler warnings about unused    \
         variables */                            \
         intp nowarns = ss1+sb1+*x1;             \
         char *x2 = mem[arg2];                   \
         intp ss2 = params.memsizes[arg2];       \
-        intp sb2 = params.memsteps[arg2];       \
+        intp sb2 = memsteps[arg2];              \
         char *x3 = mem[arg3];                   \
         intp ss3 = params.memsizes[arg3];       \
-        intp sb3 = params.memsteps[arg3];       \
+        intp sb3 = memsteps[arg3];              \
         nowarns += ss2+sb2+*x2;                 \
         nowarns += ss3+sb3+*x3;                 \
         VEC_LOOP(expr);                         \
@@ -95,37 +114,19 @@
         expr;                                   \
     } break
 
-
-    unsigned int pc, j, k, r;
+    unsigned int pc, j;
 
     /* set up pointers to next block of inputs and outputs */
-    mem[0] = params.output + index * params.memsteps[0];
-    for (r = 0; r < params.n_inputs; r++) {
-        struct index_data id = params.index_data[r+1];
-        if (id.count) {
-            mem[1+r] = params.inputs[r];
-            for (j = 0; j < BLOCK_SIZE; j++) {
-                unsigned int flatindex = 0;
-                for (k = 0; k < id.count; k ++)
-                    flatindex += id.strides[k] * id.index[k];
-                memcpy(mem[1+r] + (j*id.size), id.buffer + flatindex, id.size);
-                k = id.count - 1;
-                id.index[k] += 1;
-                if (id.index[k] >= id.shape[k])
-                    while (id.index[k] >= id.shape[k]) {
-                        id.index[k] -= id.shape[k];
-                        if (k < 1) break;
-                        id.index[k-1] += 1;
-                        k -= 1;
-                    }
-            }
-        } else {
-            mem[1+r] = params.inputs[r] + index * params.memsteps[1+r];
-        }
-    }
+#ifdef SINGLE_ITEM_CONST_LOOP
+    mem[0] = params.output;
+#else
+    /* use the iterator's inner loop data */
+    memcpy(mem, iter_dataptr, (1+params.n_inputs)*sizeof(char*));
+    memcpy(memsteps, iter_strides, (1+params.n_inputs)*sizeof(intp));
+#endif
 
     /* WARNING: From now on, only do references to mem[arg[123]]
-       & params.memsteps[arg[123]] inside the VEC_ARG[123] macros,
+       & memsteps[arg[123]] inside the VEC_ARG[123] macros,
        or you will risk accessing invalid addresses.  */
 
     for (pc = 0; pc < params.prog_len; pc += 4) {
@@ -134,14 +135,22 @@
         unsigned int arg1 = params.program[pc+2];
         unsigned int arg2 = params.program[pc+3];
         #define      arg3   params.program[pc+5]
-        #define store_index params.index_data[store_in]
-        #define reduce_ptr  (dest + flat_index(&store_index, j))
-        #define i_reduce    *(int *)reduce_ptr
-        #define l_reduce    *(long long *)reduce_ptr
-        #define f_reduce    *(float *)reduce_ptr
-        #define d_reduce    *(double *)reduce_ptr
-        #define cr_reduce   *(double *)ptr
-        #define ci_reduce   *((double *)ptr+1)
+        /* Iterator reduce macros */
+#ifdef REDUCTION_INNER_LOOP /* Reduce is the inner loop */
+        #define i_reduce    *(int *)dest
+        #define l_reduce    *(long long *)dest
+        #define f_reduce    *(float *)dest
+        #define d_reduce    *(double *)dest
+        #define cr_reduce   *(double *)dest
+        #define ci_reduce   *((double *)dest+1)
+#else /* Reduce is the outer loop */
+        #define i_reduce    i_dest
+        #define l_reduce    l_dest
+        #define f_reduce    f_dest
+        #define d_reduce    d_dest
+        #define cr_reduce   cr_dest
+        #define ci_reduce   ci_dest
+#endif
         #define b_dest ((char *)dest)[j]
         #define i_dest ((int *)dest)[j]
         #define l_dest ((long long *)dest)[j]
@@ -149,7 +158,7 @@
         #define d_dest ((double *)dest)[j]
         #define cr_dest ((double *)dest)[2*j]
         #define ci_dest ((double *)dest)[2*j+1]
-        #define s_dest ((char *)dest + j*params.memsteps[store_in])
+        #define s_dest ((char *)dest + j*memsteps[store_in])
         #define b1    ((char   *)(x1+j*sb1))[0]
         #define i1    ((int    *)(x1+j*sb1))[0]
         #define l1    ((long long *)(x1+j*sb1))[0]
@@ -177,7 +186,6 @@
         /* Some temporaries */
         double da, db;
         cdouble ca, cb;
-        char *ptr;
 
         switch (op) {
 
@@ -194,6 +202,7 @@
         case OP_COPY_DD: VEC_ARG1(memcpy(&d_dest, s1, sizeof(double)));
         case OP_COPY_CC: VEC_ARG1(memcpy(&cr_dest, s1, sizeof(double)*2));
 
+        /* Bool */
         case OP_INVERT_BB: VEC_ARG1(b_dest = !b1);
         case OP_AND_BBB: VEC_ARG2(b_dest = (b1 && b2));
         case OP_OR_BBB: VEC_ARG2(b_dest = (b1 || b2));
@@ -202,6 +211,7 @@
         case OP_NE_BBB: VEC_ARG2(b_dest = (b1 != b2));
         case OP_WHERE_BBBB: VEC_ARG3(b_dest = b1 ? b2 : b3);
 
+        /* Comparisons */
         case OP_GT_BII: VEC_ARG2(b_dest = (i1 > i2));
         case OP_GE_BII: VEC_ARG2(b_dest = (i1 >= i2));
         case OP_EQ_BII: VEC_ARG2(b_dest = (i1 == i2));
@@ -227,8 +237,9 @@
         case OP_EQ_BSS: VEC_ARG2(b_dest = (stringcmp(s1, s2, ss1, ss2) == 0));
         case OP_NE_BSS: VEC_ARG2(b_dest = (stringcmp(s1, s2, ss1, ss2) != 0));
 
+        /* Int */
         case OP_CAST_IB: VEC_ARG1(i_dest = (int)(b1));
-        case OP_ONES_LIKE_I: VEC_ARG1(i_dest = 1);
+        case OP_ONES_LIKE_II: VEC_ARG0(i_dest = 1);
         case OP_NEG_II: VEC_ARG1(i_dest = -i1);
 
         case OP_ADD_III: VEC_ARG2(i_dest = i1 + i2);
@@ -240,8 +251,9 @@
 
         case OP_WHERE_IBII: VEC_ARG3(i_dest = b1 ? i2 : i3);
 
+        /* Long */
         case OP_CAST_LI: VEC_ARG1(l_dest = (long long)(i1));
-        case OP_ONES_LIKE_L: VEC_ARG1(l_dest = 1);
+        case OP_ONES_LIKE_LL: VEC_ARG0(l_dest = 1);
         case OP_NEG_LL: VEC_ARG1(l_dest = -l1);
 
         case OP_ADD_LLL: VEC_ARG2(l_dest = l1 + l2);
@@ -253,10 +265,10 @@
 
         case OP_WHERE_LBLL: VEC_ARG3(l_dest = b1 ? l2 : l3);
 
-          /* Float */
+        /* Float */
         case OP_CAST_FI: VEC_ARG1(f_dest = (float)(i1));
         case OP_CAST_FL: VEC_ARG1(f_dest = (float)(l1));
-        case OP_ONES_LIKE_F: VEC_ARG1(f_dest = 1.0);
+        case OP_ONES_LIKE_FF: VEC_ARG0(f_dest = 1.0);
         case OP_NEG_FF: VEC_ARG1(f_dest = -f1);
 
         case OP_ADD_FFF: VEC_ARG2(f_dest = f1 + f2);
@@ -303,11 +315,11 @@
             VEC_ARG2(f_dest = functions_fff[arg3](f1, f2));
 #endif
 
-          /* Double */
+        /* Double */
         case OP_CAST_DI: VEC_ARG1(d_dest = (double)(i1));
         case OP_CAST_DL: VEC_ARG1(d_dest = (double)(l1));
         case OP_CAST_DF: VEC_ARG1(d_dest = (double)(f1));
-        case OP_ONES_LIKE_D: VEC_ARG1(d_dest = 1.0);
+        case OP_ONES_LIKE_DD: VEC_ARG0(d_dest = 1.0);
         case OP_NEG_DD: VEC_ARG1(d_dest = -d1);
 
         case OP_ADD_DDD: VEC_ARG2(d_dest = d1 + d2);
@@ -354,7 +366,7 @@
             VEC_ARG2(d_dest = functions_ddd[arg3](d1, d2));
 #endif
 
-            /* Complex */
+        /* Complex */
         case OP_CAST_CI: VEC_ARG1(cr_dest = (double)(i1);
                                   ci_dest = 0);
         case OP_CAST_CL: VEC_ARG1(cr_dest = (double)(l1);
@@ -363,8 +375,8 @@
                                   ci_dest = 0);
         case OP_CAST_CD: VEC_ARG1(cr_dest = d1;
                                   ci_dest = 0);
-        case OP_ONES_LIKE_C: VEC_ARG1(cr_dest = 1;
-                                  ci_dest = 0);
+        case OP_ONES_LIKE_CC: VEC_ARG0(cr_dest = 1;
+                                       ci_dest = 0);
         case OP_NEG_CC: VEC_ARG1(cr_dest = -c1r;
                                  ci_dest = -c1i);
 
@@ -415,20 +427,19 @@
         case OP_COMPLEX_CDD: VEC_ARG2(cr_dest = d1;
                                       ci_dest = d2);
 
+        /* Reductions */
         case OP_SUM_IIN: VEC_ARG1(i_reduce += i1);
         case OP_SUM_LLN: VEC_ARG1(l_reduce += l1);
         case OP_SUM_FFN: VEC_ARG1(f_reduce += f1);
         case OP_SUM_DDN: VEC_ARG1(d_reduce += d1);
-        case OP_SUM_CCN: VEC_ARG1(ptr = reduce_ptr;
-                                  cr_reduce += c1r;
+        case OP_SUM_CCN: VEC_ARG1(cr_reduce += c1r;
                                   ci_reduce += c1i);
 
         case OP_PROD_IIN: VEC_ARG1(i_reduce *= i1);
         case OP_PROD_LLN: VEC_ARG1(l_reduce *= l1);
         case OP_PROD_FFN: VEC_ARG1(f_reduce *= f1);
         case OP_PROD_DDN: VEC_ARG1(d_reduce *= d1);
-        case OP_PROD_CCN: VEC_ARG1(ptr = reduce_ptr;
-                                   da = cr_reduce*c1r - ci_reduce*c1i;
+        case OP_PROD_CCN: VEC_ARG1(da = cr_reduce*c1r - ci_reduce*c1i;
                                    ci_reduce = cr_reduce*c1i + ci_reduce*c1r;
                                    cr_reduce = da);
 
