@@ -1,3 +1,14 @@
+###################################################################
+#  Numexpr - Fast numerical array expression evaluator for NumPy.
+#
+#      License: MIT
+#      Author:  See AUTHORS.txt
+#
+#  See LICENSE.txt and LICENSES/*.txt for details about copyright and
+#  rights to use.
+####################################################################
+
+import __future__
 import sys
 import numpy
 
@@ -17,6 +28,7 @@ type_to_kind = expressions.type_to_kind
 kind_to_type = expressions.kind_to_type
 default_type = kind_to_type[expressions.default_kind]
 
+
 class ASTNode(object):
     """Abstract Syntax Tree node.
 
@@ -54,9 +66,7 @@ class ASTNode(object):
     def __hash__(self):
         if self.astType == 'alias':
             self = self.value
-        # Fast hash (see issue #43)
-        return ( hash(self.astType) ^ hash(self.astKind) ^
-                 hash(self.value) ^ hash(self.children) )
+        return hash((self.astType, self.astKind, self.value, self.children))
 
     def __str__(self):
         return 'AST(%s, %s, %s, %s, %s)' % (self.astType, self.astKind,
@@ -88,9 +98,8 @@ def expressionToAST(ex):
     This is necessary as ExpressionNode overrides many methods to act
     like a number.
     """
-    this_ast = ASTNode(ex.astType, ex.astKind, ex.value,
-                       [expressionToAST(c) for c in ex.children])
-    return this_ast
+    return ASTNode(ex.astType, ex.astKind, ex.value,
+                   [expressionToAST(c) for c in ex.children])
 
 
 def sigPerms(s):
@@ -150,10 +159,8 @@ def typeCompileAst(ast):
     else:
         value = ast.value
         children = ast.children
-    new_ast = ASTNode(ast.astType, ast.astKind, value,
-                           [typeCompileAst(c) for c in children])
-    return new_ast
-
+    return ASTNode(ast.astType, ast.astKind, value,
+                   [typeCompileAst(c) for c in children])
 
 class Register(object):
     """Abstraction for a register in the VM.
@@ -202,7 +209,11 @@ def stringToExpression(s, types, context):
     try:
         expressions._context.set_new_context(context)
         # first compile to a code object to determine the names
-        c = compile(s, '<expr>', 'eval')
+        if context.get('truediv', False):
+            flags = __future__.division.compiler_flag
+        else:
+            flags = 0
+        c = compile(s, '<expr>', 'eval', flags)
         # make VariableNode's for the names
         names = {}
         for name in c.co_names:
@@ -241,7 +252,10 @@ def getInputOrder(ast, input_order=None):
 
     if input_order:
         if variable_names != set(input_order):
-            raise ValueError("input names don't match those found in expression")
+            raise ValueError(
+                "input names (%s) don't match those found in expression (%s)"
+                % (input_order, variable_names))
+
         ordered_names = input_order
     else:
         ordered_names = list(variable_names)
@@ -310,28 +324,34 @@ def collapseDuplicateSubtrees(ast):
     for a in aliases:
         while a.value.astType == 'alias':
             a.value = a.value.value
-        a.reg = a.value.reg
+    return aliases
 
 def optimizeTemporariesAllocation(ast):
     """Attempt to minimize the number of temporaries needed, by
     reusing old ones.
     """
-    nodes = list(x for x in ast.postorderWalk() if x.reg.temporary)
+    nodes = [n for n in ast.postorderWalk() if n.reg.temporary]
     users_of = dict((n.reg, set()) for n in nodes)
+
+    node_regs = dict((n, set(c.reg for c in n.children if c.reg.temporary))
+                     for n in nodes)
     if nodes and nodes[-1] is not ast:
-        for c in ast.children:
-            if c.reg.temporary:
-                users_of[c.reg].add(ast)
-    for n in reversed(nodes):
+        nodes_to_check = nodes + [ast]
+    else:
+        nodes_to_check = nodes
+    for n in nodes_to_check:
         for c in n.children:
             if c.reg.temporary:
                 users_of[c.reg].add(n)
-    unused = {'bool' : set(), 'int' : set(), 'long': set(), 'float' : set(),
-              'double': set(), 'complex' : set(), 'str': set()}
+
+    unused = {'bool': set(), 'int': set(), 'long': set(), 'float': set(),
+              'double': set(), 'complex': set(), 'str': set()}
     for n in nodes:
-        for reg, users in users_of.iteritems():
-            if n in users:
-                users.remove(n)
+        for c in n.children:
+            reg = c.reg
+            if reg.temporary:
+                users = users_of[reg]
+                users.discard(n)
                 if not users:
                     unused[reg.node.astKind].add(reg)
         if unused[n.astKind]:
@@ -361,7 +381,7 @@ def setRegisterNumbersForTemporaries(ast, start):
             node.reg.n = node.value
             continue
         reg = node.reg
-        if reg.n < 0:
+        if reg.n is None:
             reg.n = start + seen
             seen += 1
             signature += reg.node.typecode()
@@ -378,13 +398,8 @@ def convertASTtoThreeAddrForm(ast):
     I suppose this should be called three register form, but three
     address form is found in compiler theory.
     """
-    program = []
-    for node in ast.allOf('op'):
-        children = node.children
-        instr = (node.value, node.reg) \
-                + tuple([c.reg for c in children])
-        program.append(instr)
-    return program
+    return [(node.value, node.reg) + tuple([c.reg for c in node.children])
+            for node in ast.allOf('op')]
 
 def compileThreeAddrForm(program):
     """Given a three address form of the program, compile it a string that
@@ -394,7 +409,7 @@ def compileThreeAddrForm(program):
         if reg is None:
             return '\xff'
         elif reg.n < 0:
-            raise ValueError("negative value for register number %s" % (reg.n,))
+            raise ValueError("negative value for register number %s" % reg.n)
         else:
             return chr(reg.n)
 
@@ -405,10 +420,10 @@ def compileThreeAddrForm(program):
         ca2 = nToChr(a2)
         return cop + cs + ca1 + ca2
 
-    def toString(*args):
+    def toString(args):
         while len(args) < 4:
             args += (None,)
-        opcode, store, a1, a2 = args[0:4]
+        opcode, store, a1, a2 = args[:4]
         s = quadrupleToString(opcode, store, a1, a2)
         l = [s]
         args = args[4:]
@@ -418,31 +433,38 @@ def compileThreeAddrForm(program):
             args = args[3:]
         return ''.join(l)
 
-    prog_str = ''.join([toString(*t) for t in program])
+    prog_str = ''.join([toString(t) for t in program])
     return prog_str
 
 context_info = [
     ('optimization', ('none', 'moderate', 'aggressive'), 'aggressive'),
+    ('truediv', (False, True, 'auto'), 'auto')
                ]
-def getContext(map):
+
+def getContext(kwargs, frame_depth=1):
+    d = kwargs.copy()
     context = {}
     for name, allowed, default in context_info:
-        value = map.pop(name, default)
+        value = d.pop(name, default)
         if value in allowed:
             context[name] = value
         else:
             raise ValueError("'%s' must be one of %s" % (name, allowed))
-    if map:
-        raise ValueError("Unknown keyword argument '%s'" % map.popitem()[0])
+
+    if d:
+        raise ValueError("Unknown keyword argument '%s'" % d.popitem()[0])
+    if context['truediv'] == 'auto':
+        caller_globals = sys._getframe(frame_depth + 1).f_globals
+        context['truediv'] = \
+            caller_globals.get('division', None) == __future__.division
+
     return context
 
-
-def precompile(ex, signature=(), copy_args=(), **kwargs):
+def precompile(ex, signature=(), context={}):
     """Compile the expression to an intermediate form.
     """
     types = dict(signature)
     input_order = [name for (name, type_) in signature]
-    context = getContext(kwargs)
 
     if isinstance(ex, str):
         ex = stringToExpression(ex, types, context)
@@ -452,29 +474,20 @@ def precompile(ex, signature=(), copy_args=(), **kwargs):
 
     ast = expressionToAST(ex)
 
-    # Add a copy for strided or unaligned unidimensional arrays
-    for a in ast.postorderWalk():
-        if a.astType == "variable" and a.value in copy_args:
-            newVar = ASTNode(*a.key())
-            a.astType, a.value, a.children = ('op', 'copy', (newVar,))
-
-    if ex.astType not in ('op'):
+    if ex.astType != 'op':
         ast = ASTNode('op', value='copy', astKind=ex.astKind, children=(ast,))
 
     ast = typeCompileAst(ast)
 
-    reg_num = [-1]
-    def registerMaker(node, temporary=False):
-        reg = Register(node, temporary=temporary)
-        reg.n = reg_num[0]
-        reg_num[0] -= 1
-        return reg
+    aliases = collapseDuplicateSubtrees(ast)
 
     assignLeafRegisters(ast.allOf('raw'), Immediate)
-    assignLeafRegisters(ast.allOf('variable', 'constant'), registerMaker)
-    assignBranchRegisters(ast.allOf('op'), registerMaker)
+    assignLeafRegisters(ast.allOf('variable', 'constant'), Register)
+    assignBranchRegisters(ast.allOf('op'), Register)
 
-    collapseDuplicateSubtrees(ast)
+    # assign registers for aliases
+    for a in aliases:
+        a.reg = a.value.reg
 
     input_order = getInputOrder(ast, input_order)
     constants_order, constants = getConstants(ast)
@@ -511,8 +524,21 @@ def NumExpr(ex, signature=(), copy_args=(), **kwargs):
 
     Returns a `NumExpr` object containing the compiled function.
     """
+    # NumExpr can be called either directly by the end-user, in which case
+    # kwargs need to be sanitized by getContext, or by evaluate,
+    # in which case kwargs are in already sanitized.
+    # In that case frame_depth is wrong (it should be 2) but it doesn't matter
+    # since it will not be used (because truediv='auto' has already been
+    # translated to either True or False).
+
+    # NOTE: `copy_args` is not necessary from 2.0 on.  It remains here
+    # basically because PyTables trusted on it for certain operations.
+    # I have filed a ticket for PyTables asking for its removal:
+    # https://github.com/PyTables/PyTables/issues/117
+
+    context = getContext(kwargs, frame_depth=1)
     threeAddrProgram, inputsig, tempsig, constants, input_names = \
-                      precompile(ex, signature, copy_args, **kwargs)
+                      precompile(ex, signature, context)
     program = compileThreeAddrForm(threeAddrProgram)
     return interpreter.NumExpr(inputsig, tempsig, program, constants,
                                input_names)
@@ -608,20 +634,56 @@ def getExprNames(text, context):
 _names_cache = CacheDict(256)
 _numexpr_cache = CacheDict(256)
 
-def evaluate(ex, local_dict=None, global_dict=None, **kwargs):
-    """Evaluate a simple array expression element-wise.
+def evaluate(ex, local_dict=None, global_dict=None,
+             out=None, order='K', casting='safe', **kwargs):
+    """Evaluate a simple array expression element-wise, using the new iterator.
 
     ex is a string forming an expression, like "2*a+3*b". The values for "a"
     and "b" will by default be taken from the calling function's frame
     (through use of sys._getframe()). Alternatively, they can be specifed
     using the 'local_dict' or 'global_dict' arguments.
+
+    Parameters
+    ----------
+
+    local_dict : dictionary, optional
+        A dictionary that replaces the local operands in current frame.
+
+    global_dict : dictionary, optional
+        A dictionary that replaces the global operands in current frame.
+
+    out : NumPy array, optional
+        An existing array where the outcome is going to be stored.  Care is
+        required so that this array has the same shape and type than the
+        actual outcome of the computation.  Useful for avoiding unnecessary
+        new array allocations.
+
+    order : {'C', 'F', 'A', or 'K'}, optional
+        Controls the iteration order for operands. 'C' means C order, 'F'
+        means Fortran order, 'A' means 'F' order if all the arrays are
+        Fortran contiguous, 'C' order otherwise, and 'K' means as close to
+        the order the array elements appear in memory as possible.  For
+        efficient computations, typically 'K'eep order (the default) is
+        desired.
+
+    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
+        Controls what kind of data casting may occur when making a copy or
+        buffering.  Setting this to 'unsafe' is not recommended, as it can
+        adversely affect accumulations.
+
+          * 'no' means the data types should not be cast at all.
+          * 'equiv' means only byte-order changes are allowed.
+          * 'safe' means only casts which can preserve values are allowed.
+          * 'same_kind' means only safe casts or casts within a kind,
+            like float64 to float32, are allowed.
+          * 'unsafe' means any data conversions may be done.
     """
     if not isinstance(ex, str):
         raise ValueError("must specify expression as a string")
     # Get the names for this expression
-    expr_key = (ex, tuple(sorted(kwargs.items())))
+    context = getContext(kwargs, frame_depth=1)
+    expr_key = (ex, tuple(sorted(context.items())))
     if expr_key not in _names_cache:
-        context = getContext(kwargs)
         _names_cache[expr_key] = getExprNames(ex, context)
     names, ex_uses_vml = _names_cache[expr_key]
     # Get the arguments based on the names.
@@ -630,61 +692,25 @@ def evaluate(ex, local_dict=None, global_dict=None, **kwargs):
         local_dict = call_frame.f_locals
     if global_dict is None:
         global_dict = call_frame.f_globals
-    arguments = []
-    copy_args = []
 
+    arguments = []
     for name in names:
         try:
             a = local_dict[name]
         except KeyError:
             a = global_dict[name]
-        b = numpy.asarray(a)
-        # Byteswapped arrays are dealt with in the extension
-        # All the opcodes can deal with strided arrays directly as
-        # long as they are undimensional (strides in other
-        # dimensions are dealt within the extension), so we don't
-        # need a copy for the strided case.
-
-        if not b.flags.aligned:
-            # Only take actions if CPU is different from AMD and Intel
-            # as they can deal with unaligned arrays very efficiently.
-            # If using VML, do the copy as the VML functions works
-            # much faster with aligned arrays.
-            if not is_cpu_amd_intel or use_vml:
-                # For the unaligned case, we have two cases:
-                if b.ndim == 1:
-                    # For unidimensional arrays we can use the copy
-                    # opcode because it can deal with unaligned arrays
-                    # as long as they are unidimensionals with a
-                    # possible stride (very common case for
-                    # recarrays).  This can be up to 2x faster than
-                    # doing a copy using NumPy.
-                    copy_args.append(name)
-                else:
-                    # For multimensional unaligned arrays do a plain
-                    # copy.  We could refine more this and do a plain
-                    # copy only in the case that strides doesn't exist
-                    # in dimensions other than the last one (whose
-                    # case is supported by the copy opcode).
-                    b = b.copy()
-        elif use_vml and ex_uses_vml: #only make a copy of strided arrays if
-                                      #vml is in use
-            if not b.flags.contiguous:
-                if b.ndim == 1:
-                    copy_args.append(name)
-                else:
-                    b = b.copy()
-
-        arguments.append(b)
+        arguments.append(numpy.asarray(a))
 
     # Create a signature
     signature = [(name, getType(arg)) for (name, arg) in zip(names, arguments)]
-    # Look up numexpr if possible. copy_args *must* be added to the key,
-    # just in case a non-copy expression is already in cache.
-    numexpr_key = expr_key + (tuple(signature),) + tuple(copy_args)
+
+    # Look up numexpr if possible.
+    numexpr_key = expr_key + (tuple(signature),)
     try:
         compiled_ex = _numexpr_cache[numexpr_key]
     except KeyError:
         compiled_ex = _numexpr_cache[numexpr_key] = \
-                      NumExpr(ex, signature, copy_args, **kwargs)
-    return compiled_ex(*arguments)
+                      NumExpr(ex, signature, **context)
+    kwargs = {'out': out, 'order': order, 'casting': casting,
+              'ex_uses_vml': ex_uses_vml}
+    return compiled_ex(*arguments, **kwargs)
