@@ -1,4 +1,15 @@
+###################################################################
+#  Numexpr - Fast numerical array expression evaluator for NumPy.
+#
+#      License: MIT
+#      Author:  See AUTHORS.txt
+#
+#  See LICENSE.txt and LICENSES/*.txt for details about copyright and
+#  rights to use.
+####################################################################
+
 import new, sys, os
+import warnings
 
 import numpy
 from numpy import (
@@ -9,7 +20,7 @@ from numpy import (
     sinh, cosh, tanh, arcsinh, arccosh, arctanh,
     log, log1p, log10, exp, expm1)
 from numpy.testing import *
-from numpy import shape, allclose, ravel, isnan, isinf
+from numpy import shape, allclose, array_equal, ravel, isnan, isinf
 
 import numexpr
 from numexpr import E, NumExpr, evaluate, disassemble, use_vml
@@ -19,8 +30,9 @@ TestCase = unittest.TestCase
 
 double = numpy.double
 
+
 # Recommended minimum versions
-minimum_numpy_version = "1.2"
+minimum_numpy_version = "1.6"
 
 class test_numexpr(TestCase):
 
@@ -73,29 +85,42 @@ class test_numexpr(TestCase):
                       ('prod_ddn', 'r0', 't3', 2)])
         # Check that full reductions work.
         x = zeros(1e5)+.01   # checks issue #41
-        assert_equal(evaluate("sum(x+2,axis=0)"), sum(x+2,axis=0))
-        assert_equal(evaluate("prod(x,axis=0)"), prod(x,axis=0))
+        assert_allclose(evaluate("sum(x+2,axis=None)"), sum(x+2,axis=None))
+        assert_allclose(evaluate("sum(x+2,axis=0)"), sum(x+2,axis=0))
+        assert_allclose(evaluate("prod(x,axis=0)"), prod(x,axis=0))
+
+        x = arange(10.0)
+        assert_allclose(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
+        assert_allclose(evaluate("prod(x**2+2,axis=0)"), prod(x**2+2,axis=0))
+
+        x = arange(100.0)
+        assert_allclose(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
+        assert_allclose(evaluate("prod(x-1,axis=0)"), prod(x-1,axis=0))
+        x = linspace(0.1,1.0,2000)
+        assert_allclose(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
+        assert_allclose(evaluate("prod(x-1,axis=0)"), prod(x-1,axis=0))
+
         # Check that reductions along an axis work
         y = arange(9.0).reshape(3,3)
-        assert_equal(evaluate("sum(y**2, axis=1)"), sum(y**2, axis=1))
-        assert_equal(evaluate("sum(y**2, axis=0)"), sum(y**2, axis=0))
-        assert_equal(evaluate("sum(y**2, axis=None)"), sum(y**2, axis=None))
-        assert_equal(evaluate("prod(y**2, axis=1)"), prod(y**2, axis=1))
-        assert_equal(evaluate("prod(y**2, axis=0)"), prod(y**2, axis=0))
-        assert_equal(evaluate("prod(y**2, axis=None)"), prod(y**2, axis=None))
+        assert_allclose(evaluate("sum(y**2, axis=1)"), sum(y**2, axis=1))
+        assert_allclose(evaluate("sum(y**2, axis=0)"), sum(y**2, axis=0))
+        assert_allclose(evaluate("sum(y**2, axis=None)"), sum(y**2, axis=None))
+        assert_allclose(evaluate("prod(y**2, axis=1)"), prod(y**2, axis=1))
+        assert_allclose(evaluate("prod(y**2, axis=0)"), prod(y**2, axis=0))
+        assert_allclose(evaluate("prod(y**2, axis=None)"), prod(y**2, axis=None))
         # Check integers
         x = arange(10.)
         x = x.astype(int)
-        assert_equal(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
-        assert_equal(evaluate("prod(x**2+2,axis=0)"), prod(x**2+2,axis=0))
+        assert_allclose(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
+        assert_allclose(evaluate("prod(x**2+2,axis=0)"), prod(x**2+2,axis=0))
         # Check longs
         x = x.astype(long)
-        assert_equal(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
-        assert_equal(evaluate("prod(x**2+2,axis=0)"), prod(x**2+2,axis=0))
+        assert_allclose(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
+        assert_allclose(evaluate("prod(x**2+2,axis=0)"), prod(x**2+2,axis=0))
         # Check complex
-        x = x + 5j
-        assert_equal(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
-        assert_equal(evaluate("prod(x**2+2,axis=0)"), prod(x**2+2,axis=0))
+        x = x + .1j
+        assert_allclose(evaluate("sum(x**2+2,axis=0)"), sum(x**2+2,axis=0))
+        assert_allclose(evaluate("prod(x-1,axis=0)"), prod(x-1,axis=0))
 
     def test_axis(self):
         y = arange(9.0).reshape(3,3)
@@ -111,12 +136,16 @@ class test_numexpr(TestCase):
             pass
         else:
             raise ValueError("should raise exception!")
-
-
-
+        try:
+            # Negative axis are not supported
+            evaluate("sum(y, axis=-1)")
+        except ValueError:
+            pass
+        else:
+            raise ValueError("should raise exception!")
 
     def test_r0_reuse(self):
-        assert_equal(disassemble(NumExpr("x**2+2", [('x', double)])),
+        assert_equal(disassemble(NumExpr("x * x + 2", [('x', double)])),
                     [('mul_ddd', 'r0', 'r1[x]', 'r1[x]'),
                      ('add_ddd', 'r0', 'r0', 'c2[2.0]')])
 
@@ -156,6 +185,25 @@ class test_evaluate(TestCase):
         x2[1] = 1
         assert_array_equal(x2, y)
 
+    # Test for issue #22
+    def test_true_div(self):
+        x = arange(10, dtype='i4')
+        assert_array_equal(evaluate("x/2"), x / 2)
+        assert_array_equal(evaluate("x/2", truediv=False), x / 2)
+        assert_array_equal(evaluate("x/2", truediv='auto'), x / 2)
+        assert_array_equal(evaluate("x/2", truediv=True), x / 2.0)
+
+    # PyTables uses __nonzero__ among ExpressionNode objects internally
+    # so this should be commented out for the moment.  See #24.
+    def _test_boolean_operator(self):
+        x = arange(10, dtype='i4')
+        try:
+            evaluate("(x > 1) and (x < 9)")
+        except TypeError:
+            pass
+        else:
+            raise ValueError("should raise exception!")
+
     def test_rational_expr(self):
         a = arange(1e6)
         b = arange(1e6) * 0.1
@@ -201,7 +249,7 @@ class test_evaluate(TestCase):
     def test_all_scalar(self):
         a = 3.
         b = 4.
-        assert_equal(evaluate("a+b"), a+b)
+        assert_allclose(evaluate("a+b"), a+b)
         expr = NumExpr("2*a+3*b",[('a', double),('b', double)])
         assert_equal(expr(a,b), 2*a+3*b)
 
@@ -255,7 +303,7 @@ tests = [
           '1+1',
           '1',
           'cos(a2)',
-          '(a+1)**0'])]
+          ])]
 
 optests = []
 for op in list('+-*/%') + ['**']:
@@ -282,31 +330,35 @@ for func in ['copy', 'ones_like', 'sqrt',
              'sinh', 'cosh', 'tanh', 'arcsinh', 'arccosh', 'arctanh',
              'log', 'log1p', 'log10', 'exp', 'expm1', 'abs']:
     func1tests.append("a + %s(b+c)" % func)
-tests.append(('1-ARG FUNCS', func1tests))
+tests.append(('1_ARG_FUNCS', func1tests))
 
 func2tests = []
 for func in ['arctan2', 'fmod']:
     func2tests.append("a + %s(b+c, d+1)" % func)
     func2tests.append("a + %s(b+c, 1)" % func)
     func2tests.append("a + %s(1, d+1)" % func)
-tests.append(('2-ARG FUNCS', func2tests))
+tests.append(('2_ARG_FUNCS', func2tests))
 
 powtests = []
-for n in (-2.5, -1.5, -1.3, -.5, 0, 0.5, 1, 0.5, 1, 2.3, 2.5):
+# n = -1, 0.5, 2, 4 already handled in section "OPERATIONS"
+for n in (-7, -2.5, -1.5, -1.3, -.5, 0, 0.0, 1, 2.3, 2.5, 3):
     powtests.append("(a+1)**%s" % n)
-tests.append(('POW TESTS', powtests))
+tests.append(('POW_TESTS', powtests))
 
 def equal(a, b, exact):
+    if array_equal(a, b):
+        return True
+
     if hasattr(a, 'dtype') and a.dtype in ['f4','f8']:
         nnans = isnan(a).sum()
-        if isnan(a).sum() > 0:
+        if nnans > 0:
             # For results containing NaNs, just check that the number
             # of NaNs is the same in both arrays.  This check could be
             # made more exhaustive, but checking element by element in
             # python space is very expensive in general.
             return nnans == isnan(b).sum()
         ninfs = isinf(a).sum()
-        if isinf(a).sum() > 0:
+        if ninfs > 0:
             # Ditto for Inf's
             return ninfs == isinf(b).sum()
     if exact:
@@ -324,22 +376,32 @@ class Skip(Exception): pass
 def test_expressions():
     test_no = [0]
     def make_test_method(a, a2, b, c, d, e, x, expr,
-                         test_scalar, dtype, optimization, exact):
+                         test_scalar, dtype, optimization, exact, section):
         this_locals = locals()
         def method():
-            npval = eval(expr, globals(), this_locals)
+            # We don't want to listen at RuntimeWarnings like
+            # "overflows" or "divide by zero".  Feel free to expand
+            # the range for this filter, if needed.
+            if dtype.__name__ == "float32" or 'arctanh' in expr:
+                warnings.simplefilter("ignore")
+                npval = eval(expr, globals(), this_locals)
+                warnings.simplefilter("always")
+            else:
+                npval = eval(expr, globals(), this_locals)
             try:
                 neval = evaluate(expr, local_dict=this_locals,
                                  optimization=optimization)
                 assert equal(npval, neval, exact), """%r
 (test_scalar=%r, dtype=%r, optimization=%r, exact=%r,
- npval=%r (%r)\n neval=%r (%r))""" % (expr, test_scalar, dtype.__name__,
+ npval=%r (%r - %r)\n neval=%r (%r - %r))""" % (expr, test_scalar, dtype.__name__,
                                      optimization, exact,
-                                     npval, type(npval), neval, type(neval))
+                                     npval, type(npval), shape(npval),
+                                     neval, type(neval), shape(neval))
             except AssertionError:
                 raise
             except NotImplementedError:
-                print('%r not implemented for %s' % (expr,dtype.__name__))
+                print('%r not implemented for %s (scalar=%d, opt=%s)'
+                      % (expr, dtype.__name__, test_scalar, optimization))
             except:
                 print('numexpr error for expression %r' % (expr,))
                 raise
@@ -347,11 +409,15 @@ def test_expressions():
                               'dtype=%r, optimization=%r, exact=%r)') \
                     % (expr, test_scalar, dtype.__name__, optimization, exact)
         test_no[0] += 1
-        method.__name__ = 'test_%04d' % (test_no[0],)
+        method.__name__ = 'test_scalar%d_%s_%s_%s_%04d' % (test_scalar,
+                                                           dtype.__name__,
+                                                           optimization,
+                                                           section,
+                                                           test_no[0])
         return method
     x = None
-    for test_scalar in [0,1,2]:
-        for dtype in [int, long, numpy.float32, double, complex]:
+    for test_scalar in (0, 1, 2):
+        for dtype in (int, long, numpy.float32, double, complex):
             array_size = 100
             a = arange(2*array_size, dtype=dtype)[::2]
             a2 = zeros([array_size, array_size], dtype=dtype)
@@ -365,9 +431,9 @@ def test_expressions():
                     x += 1j
                     x *= 1+1j
             if test_scalar == 1:
-                a = a[array_size/2]
+                a = a[array_size / 2]
             if test_scalar == 2:
-                b = b[array_size/2]
+                b = b[array_size / 2]
             for optimization, exact in [
                 ('none', False), ('moderate', False), ('aggressive', False)]:
                 for section_name, section_tests in tests:
@@ -375,16 +441,18 @@ def test_expressions():
                         if dtype == complex and (
                                '<' in expr or '>' in expr or '%' in expr
                                or "arctan2" in expr or "fmod" in expr):
-                             # skip complex comparisons or functions not
-                             # defined in complex domain.
+                            # skip complex comparisons or functions not
+                            # defined in complex domain.
                             continue
                         if (dtype in (int, long) and test_scalar and
                             expr == '(a+1) ** -1'):
                             continue
+
                         m = make_test_method(a, a2, b, c, d, e, x,
                                              expr, test_scalar, dtype,
-                                             optimization, exact)
-                        yield m,
+                                             optimization, exact,
+                                             section_name)
+                        yield m
 
 class test_int64(TestCase):
     def test_neg(self):
@@ -561,6 +629,28 @@ class test_irregular_stride(TestCase):
         assert_array_equal(f0[i0], arange(5, dtype=int32))
         assert_array_equal(f1[i1], arange(5, dtype=float64))
 
+# Cases for testing arrays with dimensions that can be zero.
+class test_zerodim(TestCase):
+
+    def test_zerodim1d(self):
+        a0 = array([], dtype=int32)
+        a1 = array([], dtype=float64)
+
+        r0 = evaluate('a0 + a1')
+        r1 = evaluate('a0 * a1')
+
+        assert_array_equal(r0, a1)
+        assert_array_equal(r1, a1)
+
+    def test_zerodim3d(self):
+        a0 = array([], dtype=int32).reshape(0,2,4)
+        a1 = array([], dtype=float64).reshape(0,2,4)
+
+        r0 = evaluate('a0 + a1')
+        r1 = evaluate('a0 * a1')
+
+        assert_array_equal(r0, a1)
+        assert_array_equal(r1, a1)
 
 # Case test for threads
 class test_threading(TestCase):
@@ -585,7 +675,10 @@ def _worker(qout = None):
 # Case test for subprocesses (via multiprocessing module)
 class test_subprocess(TestCase):
     def test_multiprocess(self):
-        import multiprocessing as mp
+        try:
+            import multiprocessing as mp
+        except ImportError:
+	    return
         # Check for two threads at least
         numexpr.set_num_threads(2)
         #print "**** Running from main process:"
@@ -640,12 +733,16 @@ def suite():
 
     class TestExpressions(TestCase):
         pass
-    for m, in test_expressions():
+
+    def add_method(func):
         def method(self):
-            return m()
-        setattr(TestExpressions, m.__name__,
+            return func()
+        setattr(TestExpressions, func.__name__,
                 new.instancemethod(method, None, TestExpressions))
 
+    for func in test_expressions():
+        add_method(func)
+
     for n in range(niter):
         theSuite.addTest(unittest.makeSuite(test_numexpr1))
         theSuite.addTest(unittest.makeSuite(test_numexpr2))
@@ -656,6 +753,7 @@ def suite():
         theSuite.addTest(unittest.makeSuite(test_strings))
         theSuite.addTest(
             unittest.makeSuite(test_irregular_stride) )
+        theSuite.addTest(unittest.makeSuite(test_zerodim))
         theSuite.addTest(unittest.makeSuite(test_subprocess))
         # I need to put this test after test_subprocess because
         # if not, the test suite locks immediately before test_subproces.
@@ -668,3 +766,5 @@ def suite():
 if __name__ == '__main__':
     print_versions()
     unittest.main(defaultTest = 'suite')
+#    suite = suite()
+#    unittest.TextTestRunner(verbosity=2).run(suite)
